import time
import traceback
import os.path
import math
import logging

import psutil
import cv2
from cv2cuda.ffmpeg_process import FFMPEG
from cv2cuda.decorator import timeit

logger = logging.getLogger(__name__)
check_log = logging.getLogger(__name__ + ".check")


class FFMPEGVideoWriter:
    """
    A cv2.VideoWriter-like interface that supports FFMPEG+CUDA
    for faster and efficient encoding of videos
    """

    _TIMEOUT=3

    def __init__(self, filename, apiPreference, fourcc, fps, frameSize, isColor=False, maxframes=math.inf, yes=True, device="gpu"):

        self._isColor = isColor # color not supported for now
        self._fourcc = fourcc
        self._apiPreference = apiPreference
        if " " in filename:
            logger.warning("Please dont pass spaces in the filename, as the video writer may not work")
        self._filename = filename
        self._fps = fps
        self._frameSize = frameSize
        width, height = frameSize # wxh
        self._terminate_time = None
        self._count = 0
        self._maxframes = maxframes
        self._already_warned = False

        self._check_deps()


        if os.path.exists(filename):
            if not yes:
                answer = input(f"{filename} already exists, overwrite? Y/n")
                if answer.lower() == "y":
                    pass
                else:
                    raise Exception(f"{filename} exists")
            else:
                logger.warning(f"{filename} exists already. Overwriting.")


        if isColor:
            raise Exception(
                f"""User supplied isColor={isColor}.
                cv2cuda.VideoWriter does not support color.
                Terminating ...
                """
            )
                
        if maxframes is not math.inf:
            logger.warning(
                f"""
                User supplied maxframes={maxframes}.
                Program calling cv2cuda.VideoWriter may malfunction after {maxframes} are captured
                """
            )

        if device != "gpu":
            logger.warning(
                f"""User supplied device={device}.
                The GPU and CUDA drivers will NOT be used.
                """
            )

        self._ffmpeg = FFMPEG(width, height=height, fps=fps, output=filename, device=device, codec=fourcc, encode=True)


    @timeit
    def write(self, image):
        if len(image.shape) == 3:
            if not self._already_warned:
                logger.warning(
                    """
                    Color frames are not supported, please provide gray images to remove this warning
                    I will force the frames gray now, which may add computational time which could be spared
                    """
                )
                self._already_warned = True

            image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        self._ffmpeg.write(image)
        self._count += 1
        if self._count == self._maxframes:
            self.release()


    def _check_cuda(self):
        logger.warning("CUDA checks not implemented yet")
        return
    

    def _check_ffmpeg(self):
        logger.warning("FFMPEG checks not implemented yet")
        return


    def _check_deps(self):
        pass
        # self._check_ffmpeg()
        # self._check_cuda()

    
    def _check_terminated(self):

        check_log.debug(self._ffmpeg._command)
        check_log.debug(self._count)

        for process in psutil.process_iter():
            process_line = " ".join(process.cmdline())            
            if self._ffmpeg._command == process_line:
                msg = f"{self._ffmpeg._command} is stuck"
                # assert False, msg
                check_log.warning(msg)
                return False
        
        return True

    def release(self):
        code = self._ffmpeg.terminate()
        logger.info(f"Termination code for {self._ffmpeg._command}: {code}")
        self._terminate_time = time.time()
        try:
            status = self._ffmpeg.wait(self._TIMEOUT)
        except Exception as error:
            logger.error(error)
            logger.error(traceback.print_exc())
            self._ffmpeg.kill()
            status = 1

        check_log.debug(f"STATUS: {status}")        
        if not self._check_terminated():
            check_log.debug(f"Killing {self._ffmpeg._command}")
            self._ffmpeg.kill()
        
                
        return status


VideoWriter = FFMPEGVideoWriter


class CV2VideoWriter(cv2.VideoWriter):

    @timeit
    def write(self, *args, **kwargs):
        return super().write(*args, **kwargs)

    @timeit
    def read(self, *args, **kwargs):
        return super().read(*args, **kwargs)


from functools import wraps
from time import time

def timeit(f):
    @wraps(f)
    def wrap(*args, **kwargs):
        ts = time()
        result = f(*args, **kwargs)
        te = time()
        msec = (te-ts)*1000
        return result, msec
    
    wrap.unwrapped = f
    return wrap


import subprocess
import shlex
import logging
import threading
import time

PIX_FMT = "gray" # graycolor format

logger = logging.getLogger(__name__)
write_log = logging.getLogger(__name__ + ".write")
terminate_log = logging.getLogger(__name__ + ".terminate")
# write_log.setLevel(logging.DEBUG)
# terminate_log.setLevel(logging.DEBUG)


class FFMPEG:
    
    def __init__(self, *args, **kwargs):
        command, registers = self._setup(*args, **kwargs)
        cmd = shlex.split(command)
        self._cmd = cmd
        self._command = command
        logger.debug(cmd)

        self._process = subprocess.Popen(
            cmd,
            stdin=registers[0],
            stdout=registers[1],
            shell=False
        )
        self._terminate_event = False

        self._validate_popen()
        self._lock = threading.Lock()



    def _validate_popen(self):

        if self._process.poll() is None:
            logger.info(f"{self._command} is alive")


    def _setup(self, width, height, fps, output, device="gpu", codec="h264_nvenc", encode=True):

        if not encode:
            raise Exception("Decoder is not yet implemented")

        if device == "gpu":
            command = f"ffmpeg -y -loglevel warning -r {fps} -f rawvideo -pix_fmt {PIX_FMT}"\
                " -vsync 0 -extra_hw_frames 2"\
                f" -s {width}x{height}"       
            if output is None:
                command += f" -i - -an -c:v {codec} -f null - "
            else:
                command += f" -i - -an -c:v {codec} {output}"

            registers = (subprocess.PIPE, None)


        elif device == "cpu":
                command = f"ffmpeg -y -loglevel warning -r {fps} -f rawvideo  -pix_fmt {PIX_FMT}"\
                    f" -s {width}x{height}"
                if output is None:
                    command += f" -i - -an -vcodec {codec} -f null -"
                else:
                    command += f" -i - -an -vcodec {codec} {output}"
        
        if encode:
            registers = (subprocess.PIPE, None)
        else:
            registers = (None, subprocess.PIPE)

        return command, registers


    def write(self, image):
        if not self._terminate_event:
            with self._lock:
                try:
                    # import numpy as np
                    # image = np.random.randint(0, 255, image.shape)
                    self._process.stdin.write(image)
                    write_log.debug(f"{image.shape} to {self._command}")
                except BrokenPipeError as error:
                    write_log.warning(
                        "The FFMPEG process\n"\
                        f"{self._command}"\
                        "\nis defunct"
                    )
                    self._terminate_event = 2


    def terminate(self):
        with self._lock:
            terminate_log.debug(f"Closing standard input of {self._command}")
            self._process.stdin.close()
            self._terminate_event = True
            terminate_log.debug(f"Terminating {self._command}")
            # terminate_log.debug(f"Sleeping 10 seconds")
            # time.sleep(10)
            code = self._process.terminate()
            terminate_log.debug(f"CODE: {code}")
            return code

    
    def kill(self):
        return self._process.kill()
        

    def poll(self):
        return self._process.poll()


    def wait(self, *args, **kwargs):
        return self._process.wait(*args, **kwargs)



import logging
import os.path
import time
import threading
import multiprocessing
import math
from abc import ABC

import cv2
import cv2cuda
import cv2cuda.utils.cpu as cpu_utils
SUPPORTED_CAMERAS=["virtual", "opencv"]
# try:
#     from scicam.io.cameras import BaslerCamera #pyright: reportMissingImports=false
#     BASLER_CAMERA_ENABLED=True
# except ImportError:
#     BASLER_CAMERA_ENABLED=False

try:
    import cv2cuda.utils.gpu as gpu_utils
    GPU_PROFILING_ENABLED=True
except ImportError:
    GPU_PROFILING_ENABLED=False


def get_queue(size):
    return multiprocessing.Queue(size)


def get_camera(camera, width, height, fps, idx=0):

    if camera not in SUPPORTED_CAMERAS:
        raise Exception(f"Passed camera {camera} is not one of the supported cameras: {SUPPORTED_CAMERAS}")
    
    elif camera == "virtual":
        cap = cv2cuda.VideoCapture(idx)
        cap.set(cv2.CAP_PROP_FPS, fps)
        cap.set(cv2.CAP_PROP_FRAME_WIDTH, width)
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, height)
    
    # elif camera == "Basler":
    #     if not BASLER_CAMERA_ENABLED:
    #         raise Exception("Basler camera not supported")
    #     else:
    #         exposure = int(14000)
    #         print(exposure)

    #         cap = BaslerCamera(
    #             width=width,
    #             height=height,
    #             framerate=fps,
    #             exposure=exposure,
    #             iso=0,
    #             drop_each=1,
    #             use_wall_clock=False,
    #             timeout=30000,
    #             resolution_decrease=None,
    #             rois=None,
    #             start_time=time.time(),
    #             idx=idx
    #         )
    #         cap.open(buffersize=100)

    elif camera == "opencv":
        cap = cv2.VideoCapture(0)
        cap.set(cv2.CAP_PROP_FPS, fps)
        cap.set(cv2.CAP_PROP_FRAME_WIDTH, width)
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

    return cap


def get_video_writer(output, fps, frameSize, backend="FFMPEG", device="gpu", **kwargs):

    if device == "gpu":
        if backend == "cv2":
            raise Exception(
                "cv2.VideoWriter class using GPU is not supported in Linux."\
                "https://github.com/opencv/opencv_contrib/issues/3044"
            )
            
        elif backend == "FFMPEG":
        
            video_writer = cv2cuda.VideoWriter(
                filename = output + '.mp4',
                apiPreference="FFMPEG",
                fourcc="h264_nvenc",
                fps=fps,
                frameSize=frameSize,
                isColor=False,
                **kwargs,
            )
    elif device == "cpu":
        if backend == "cv2":
            logging.warning(
                ".avi container will be used"\
                "(instead of .mp4)"\
                "because the developer's setup does not support"\
                "encoding of videos in the cpu to .mp4"          
            )
            video_writer = cv2cuda.CV2VideoWriter(
                filename = output + '.avi',
                apiPreference=cv2.CAP_FFMPEG,
                fourcc=cv2.VideoWriter_fourcc(*"MJPG"),
                fps=fps,
                frameSize=frameSize,
                isColor=False
            )
        elif backend == "FFMPEG":
            video_writer = cv2cuda.VideoWriter(
                filename = output + '.mp4',
                apiPreference="FFMPEG",
                fourcc="mpeg4",
                fps=fps,
                frameSize=frameSize,
                isColor=False,
                **kwargs,
            )            
    
    else:
        raise Exception("device must be one of [cpu, gpu]")
    
    return video_writer



class BaseProgram(ABC):


    def __init__(self, idx, stop_queue, width, height, fps, profile, output, *args, camera="virtual", backend="FFMPEG", device="0", yes=False, duration=math.inf, **kwargs):
        self._idx = idx,
        self._stop_queue = stop_queue
        self._width = width
        self._height = height
        self._fps = fps
        self._output = output
        self._backend = backend
        self._duration = duration
        self._camera = camera
        self._yes = yes

        self._output_prefix = os.path.join(output, f"{profile}_{idx}")
       
        if profile:
            self._profile = self._output_prefix + ".profile"
        else:
            self._profile = None

        if profile:
            self._init_profile_log()
        
        self._handle = None
        self._N = None


        if device != "cpu":
            if device == "gpu": device = 0
            try:
                self._device_int = int(device)
            except ValueError:
                raise Exception("Please pass to --device either [cpu] [gpu] or a [gpu index](0, 1,..)")
            
            device = "gpu"
        else:
            self._device_int = None

        self._device = device
        super().__init__()


    @property
    def video_name(self):
        return self._output_prefix + ".mp4"

    def _init_profile_log(self):

        os.makedirs(self._profile, exist_ok=True)

        with open(self._profile, "w") as filehandle:
            filehandle.write("t,read,write,cpu_usage,encode_usage,gpu_usage\n")


    def run(self):

        cap = get_camera(self._camera, self._width, self._height, self._fps)

        video_writer = None

        start_time = time.time()


        if self._profile and GPU_PROFILING_ENABLED:
            pynvml_handles = gpu_utils.init_pynvml_handlers(self._device_int)
        else:
            pynvml_handles = None


        while (time.time() - start_time) < self._duration:
             
            if not self._stop_queue.empty() :
                self._stop_queue.get()
                logging.debug("Got STOP")
                break

            try:
                logging.debug("Reading frame")

                if "unwrapped" in dir(cap.read):
                    now = time.time()
                    if self._profile:
                        (ret, frame), read_msec = cap.read()
                    else:
                        ret, frame = cap.read.unwrapped(cap)
                else:
                    ret, frame = cap.read()


                if ret:

                    if video_writer is None:
                        video_writer = get_video_writer(
                            self.video_name, self._fps, frame.shape[:2][::-1],
                            backend=self._backend, device=self._device,
                            yes=self._yes
                        )
                    
                    logging.debug("Writing frame")
                    if self._profile:
                        _, write_msec = video_writer.write(frame)
                    else:
                        video_writer.write.unwrapped(video_writer, frame)


                    if self._profile:
                        cpu_usage = cpu_utils.query_cpu_usage()
                        if GPU_PROFILING_ENABLED:
                            enc_usage = gpu_utils.query_encoder_usage(pynvml_handles)
                            gpu_usage = gpu_utils.query_gpu_usage(pynvml_handles)
                        else:
                            enc_usage = None
                            gpu_usage = None
                        
                        with open(self._profile, "a") as filehandle:
                            data = f"{now-start_time},{read_msec},{write_msec},"\
                            f"{cpu_usage},{enc_usage},{gpu_usage}"\
                            "\n"
                            filehandle.write(data)

                else:
                    break
            
            except KeyboardInterrupt:
                pass

        logging.debug(f"Queue is empty: {self._stop_queue.qsize() == 0}")
        logging.debug("Releasing VideoCapture instance")
        cap.release()
        if video_writer:
            logging.debug("Releasing VideoWriter instance")
            video_writer.release()
        logging.debug("Process terminated")
        return

    def terminate(self):
        self._stop_queue.put("STOP")
        time.sleep(.5)
        try:
            while not self._stop_queue.empty():
                self._stop_queue.get()
            super().terminate()
        except:
            logging.debug("Terminate failed")
            pass


class Process(BaseProgram, multiprocessing.Process):
    pass

class Thread(BaseProgram, threading.Thread):
    pass



import argparse
import unittest
import tempfile
import os
import urllib.request
import cv2
from cv2cuda import VideoWriter
try:
    import progressbar # type: ignore
    PROGRESSBAR_AVAILABLE=True
except ModuleNotFoundError:
    print(
        """Install progressbar (pip install progressbar)
        if you want to visualize the progression
        of the download of the BigBuckBunny test video
        """
    )
    
    PROGRESSBAR_AVAILABLE=False

BIGBUCKBUNNY_WEB = "http://commondatastorage.googleapis.com/gtv-videos-bucket/sample/BigBuckBunny.mp4"
TEMP=False
pbar = None
DATA_DIR = "./cv2cuda/tests/static_data/"
assert os.path.exists("./cv2cuda/tests/")
os.makedirs(DATA_DIR, exist_ok=True)

def show_progress(block_num, block_size, total_size):
    global pbar
    if pbar is None and PROGRESSBAR_AVAILABLE:
        pbar = progressbar.ProgressBar(maxval=total_size)
        pbar.start()

    downloaded = block_num * block_size
    if downloaded < total_size and PROGRESSBAR_AVAILABLE:
        pbar.update(downloaded)
    elif PROGRESSBAR_AVAILABLE:
        pbar.finish()
        pbar = None


def download_file(resource, destination):
    if not os.path.exists(destination):
        print(f"Downloading {resource} -> {destination}")
        urllib.request.urlretrieve(resource, destination, show_progress)

def get_video_writer(output, fps):
    return VideoWriter(
                filename = output,
                apiPreference="FFMPEG",
                fourcc="h264_nvenc",
                fps=fps,
                frameSize=(1280, 720),
                isColor=False,
            )
   

class TestVideoWriter(unittest.TestCase):

    NFRAMES=1200
    FPS=24

    @classmethod
    def setUpClass(self):
        if TEMP:
            self._mp4_file = tempfile.NamedTemporaryFile(suffix=".mp4")
        else:
            self._mp4_file = argparse.Namespace(name=os.path.join(DATA_DIR, "BigBuckBunny.mp4"))

        download_file(BIGBUCKBUNNY_WEB, self._mp4_file.name)
        
    def setUp(self):
        if TEMP:
            self._output = tempfile.NamedTemporaryFile(suffix=".mp4")
        else:
            self._output = argparse.Namespace(name=os.path.join(DATA_DIR, "BigBuckBunny_output.mp4"))
        
        self._cap = cv2.VideoCapture(self._mp4_file.name)

    def test_VideoWriter_initializes(self):

        video_writer = get_video_writer(self._output.name, self.FPS)        

        self.assertTrue(isinstance(video_writer, VideoWriter))
        self.assertTrue("write" in dir(video_writer))
        video_writer.release()

    def test_VideoWriter_writes(self):

        video_writer = get_video_writer(self._output.name, self.FPS)        

        for i in range(self.NFRAMES):
            ret, frame = self._cap.read()
            # color data not supported
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            video_writer.write(frame)
        video_writer.release()
        
        self.assertTrue(os.path.exists(self._output.name))
        file_size = os.path.getsize(self._output.name)
        self.assertEqual(file_size, 13279390)

    def test_VideoWriter_result_plays_back(self):

        video = cv2.VideoCapture(self._output.name)
        self.assertTrue(video.get(cv2.CAP_PROP_FPS) == self.FPS)
        self.assertTrue(video.get(cv2.CAP_PROP_FRAME_COUNT) == self.NFRAMES)
        video.release()
        
    def tearDown(self):
        self._cap.release()
        # os.remove(self._output.name)
    
    @classmethod
    def tearDownClass(self):
        pass
        # os.remove(self._mp4_file.name)


if __name__ == "__main__":
    unittest.main()




"""
Run video encoding programs in parallel (either using threads or processes) to test the package
and benchmark your hardware
"""

import signal
import time
import os
import os.path

from cv2cuda.utils.components import Process, Thread, get_queue

from .parser import get_parser

signal_count = 0

def main():

    ap = get_parser()
    args = ap.parse_args()
    kwargs = vars(args)
    njobs = kwargs.pop("jobs")

    processes = [None, ] * njobs
    stop_queues = [None, ] * njobs

    if njobs == 1:
        ProgramClass = Thread
    else:
        ProgramClass = Process

    for i in range(njobs):
        stop_queues[i] = get_queue(1)
        process_kwargs = kwargs.copy()
        process_kwargs["stop_queue"] = stop_queues[i]
        processes[i] = ProgramClass(idx=i, **process_kwargs, daemon=True)

    def quitHandler(signalNumber, frame):

        global signal_count

        print(f"Received: signal.SIGINT")
        for i, process in enumerate(processes):
            process.terminate()

        signal_count += 1
        os._exit(0)

    signal.signal(signal.SIGINT, quitHandler)
    start_time = time.time()

    for i in range(njobs):
        print(i)
        processes[i].start()

    for i in range(njobs):
        print(i)
        try:
            processes[i].join()
        except KeyboardInterrupt:
            pass


if __name__ == "__main__":
    main()



import argparse

from cv2cuda.utils.components import SUPPORTED_CAMERAS

def get_parser():

    ap = argparse.ArgumentParser()
    ap.add_argument("--width", type=int, default=3860, help="width attribute of the camera and width of the recorded frames")
    ap.add_argument("--height", type=int, default=2178, help="height attribute of the camera and height of the recorded frames")
    ap.add_argument("--fps", type=int, default=30, help="Framerate of recording")
    ap.add_argument("--input", type=str, default="output", help="Input mp4")
    ap.add_argument("--output", type=str, default="output", help="Folder where output should be produced")
    ap.add_argument(
        "--camera", type=str, default="virtual",
        choices=SUPPORTED_CAMERAS
    )
    ap.add_argument("--device", default=0, help="Device to be used for encoding of video. Use cpu or gpu. An integer is understood as a GPU id")
    ap.add_argument(
        "--backend", type=str, default="FFMPEG", choices=["FFMPEG", "cv2"],
        help="""
        Backend to be used to encode the video.
        Options are cv2 (uses cv2.VideoWriter)
        or FFMPEG (calls an ffmpeg subprocess
        that reads output from the main cv2cuda process and encodes it)
        """
    )
    ap.add_argument("--jobs", type=int, default=1)
    ap.add_argument("--profile", type=str, default=None)
    ap.add_argument("--duration", type=int, default=999999)
    ap.add_argument("--yes", default=False, action="store_true")
    return ap




import unittest
import time
from cv2cuda.decorator import timeit

class TestTimeIt(unittest.TestCase):

    def setUp(self):

        @timeit
        def foo():
            time.sleep(.2)
        self._f = foo

    
    def test_timeit(self):

        self.assertTrue(self._f() is None)
        self.assertTrue(self._f.unwrapped() is None)

if __name__ == "__main__":
    unittest.main()



import pynvml as N # type: ignore

def init_pynvml_handlers(index):
    N.nvmlInit()
    handle = N.nvmlDeviceGetHandleByIndex(int())
    return N, handle

def query_gpu_usage(pynvml_handles):
    N, handle = pynvml_handles
    memory = N.nvmlDeviceGetMemoryInfo(handle)
    utilization_gpu = 100 * (memory.used / memory.total)
    return utilization_gpu

def query_encoder_usage(pynvml_handles):
    N, handle = pynvml_handles
    utilization_enc, _ = N.nvmlDeviceGetEncoderUtilization(handle)
    return utilization_enc



from setuptools import setup, find_packages

PACKAGE_NAME = "cv2cuda"

setup(
    name=PACKAGE_NAME,
    version="1.0.2",
    install_requires=["opencv-python", "numpy", "psutil"],
    packages=find_packages(),
    extras_require={"profile": ["pynvml"], "test": ["progressbar"]},
    entry_points={
        "console_scripts": [
            "cv2cuda=cv2cuda.bin.main:main"
        ]
    },
    python_requires=">=3.8.12"
)



import logging
import time

import numpy as np
import cv2

from cv2cuda.decorator import timeit

class VideoCapture:
    """
    A simulated cv2.VideoCapture class with ability to set FPS
    Useful for testing
    """

    def __init__(self, idx):
        self._idx = idx
        self._last_frame = None
        self._last_time = 0
        self._fps = 30
        self._width = None
        self._height = None


    def set(self, prop, value):

        if prop == cv2.CAP_PROP_FRAME_WIDTH:
            self._width = value
        elif prop == cv2.CAP_PROP_FRAME_HEIGHT:
            self._height = value

        elif prop == cv2.CAP_PROP_FPS:
            self._fps = value

        else:
            logging.warning("This is a simulated camera")

    def get(self, prop):
        logging.warning("This is a simulated camera")
        return None

    @timeit
    def read(self):

        assert self._height is not None
        assert self._width is not None
        assert self._fps is not None
        
        ret = True
        now = time.time()
        
        while not (now - self._last_time) > (1 / self._fps):
            time.sleep(0.001)
            now = time.time()

        try:

            self._last_frame = np.random.randint(0, 256, (self._height, self._width), np.uint8)
            self._last_time = now
            ret = True
        except:
            frame = None
            ret = False
            
        return ret, self._last_frame

    def release(self):
        return


